import 'dart:async';

import 'package:firebase_auth/firebase_auth.dart' as fb;
import 'package:firebase_core/firebase_core.dart';
import 'package:firebase_crashlytics/firebase_crashlytics.dart';
import 'package:sunny_core_widgets/sunny_core_widgets.dart';
import 'package:sunny_essentials/state.dart';
import 'package:sunny_fluro/sunny_fluro.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter_platform_widgets/flutter_platform_widgets.dart';
import 'package:logging/logging.dart';
import 'package:provider/provider.dart';
import 'package:provider/single_child_widget.dart';
import 'package:responsive_builder/responsive_builder.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_lifecycle/analytics_service.dart';
import 'package:sunny_lifecycle/app_context_events.dart';
import 'package:sunny_lifecycle/app_info.dart';
import 'package:sunny_lifecycle/app_version_service.dart';
import 'package:sunny_lifecycle/auth_state.dart';
import 'package:sunny_lifecycle/forced_upgrade.dart';
import 'package:sunny_lifecycle/help/help_service.dart';
import 'package:sunny_lifecycle/login_provider.dart';
import 'package:sunny_lifecycle/sunny_app.dart';
import 'package:sunny_lifecycle/tabs.dart';
import 'package:sunny_lifecycle/user_preferences_service.dart';
import 'package:sunny_sdk_core/auth/firebase_api_auth.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';

import 'init_interface.dart'
    if (dart.library.io) 'init_native.dart'
    if (dart.library.js) 'init_web.dart';
import 'lifecycle_extensions.dart';

final _mainLog = Logger("lifecycle");
final appErrorStream = StreamController<ErrorStack>.broadcast();
final rootNavigatorKey = GlobalKey<NavigatorState>(debugLabel: "rootNavigator");

Stream<ErrorStack> get appErrors => appErrorStream.stream;

typedef ChildWidgetBuilder = Widget Function(
    BuildContext context, Widget child);

typedef Themes ThemesBuilder(Brightness? brightness);

/// These are the variables loaded before we attempt to start the flutter runtime.
class LifecycleInit {
  /// The currently authenticated firebase user
  fb.User? get initFbUser => authState.currentFirebaseUser;

  /// The currently authenticated user details (from your system)
  UserDetails? get initProfile => authState.currentUser;

  final AuthState authState;
  final UserPreferencesService userPrefs;
  final AnalyticsService analytics;
  final FirebaseApp fbApp;
  final DeviceInfo deviceInfo;
  final FRouter router;
  final IHelpService? helpService;

  /// Your app's info, version, etc
  final AppInfo appInfo;

  LifecycleInit(
    this.deviceInfo,
    this.authState,
    this.analytics,
    this.fbApp,
    this.router,
    this.appInfo,
    this.userPrefs,
    this.helpService,
  );
}

/// This class accepts a wide array of strategies for various things, and then
/// loads the app in a way that makes sure everything is scoped properly.
///
/// It takes care of:
/// * App build versioning metadata
/// * Device/screen layouts
/// * User/authentication, displaying auth screens, logging out
/// * Push notifications (Firebase)
/// * App routing (using fluro)  This includes deep linking
/// * Analytics, crashlytics, error reporting
///
class AppBootstrap {
  static AppBootstrap? _current;

  BuildContext? _innerContext;

  set innerContext(BuildContext? context) {
    _innerContext = context;
    sunny.core.buildContext = context;
  }

  BuildContext? get innerContext => _innerContext;

  /// Stateful property that contains the parameters gathered during initialization
  LifecycleInit? lifecycleInit;

  static AppBootstrap current() {
    return _current ??
        illegalState("No app bootstrap has been invoked.  This "
            "cannot be invoked before the run method has been called");
  }

  static BuildContext currentContext() {
    return _current?.innerContext ??
        illegalState("No app bootstrap has been invoked.  This "
            "cannot be invoked before the run method has been called");
  }

  final bool isChangeBrightness;
  final String identifier;

  // final String googleApiKey;
  // final String googleSignInClientId;
  final String baseApiUrl;
  final Future<List<String>> Function() initRoutes;
  final BuildContextResolver serviceResolver;
  final TabSet tabs;

  /// You can provide this to enforce that your application is within a certain
  /// version range
  final VersionCheckSettings? versionCheckSettings;

  final Future Function(LifecycleInit? init)? completeStage1;
  final Future Function(AnalyticsService analytics) initAnalytics;
  final Widget Function(Widget inner) buildOuterContext;
  final ThemesBuilder themesBuilder;
  final void Function(BuildContext context) onUnauthenticatedError;
  final void Function(BuildContext context) initWithinApp;
  final void Function(BuildContext context)? initWithinNavigator;
  final Future<UserDetails> Function(fb.User fbUser) loadUserProfile;
  final List<LoginProvider> loginProviders;
  final List<AnalyticsProvider> analyticsProviders;
  final dynamic Function(String? type, String? body) pushMessageBodyConverter;
  final AccountCompleter<UserDetails> accountCompleter;
  final ScreenInit screenInit;
  final Widget Function(BuildContext context, Widget app) buildPlatformApp;
  final Stream<ErrorStack> appErrors;
  final Future<bool> Function()? onPreLogout;
  final Future<bool> Function(BuildContext context, fb.User user, String token)
      sendDevicePushToken;
  final FRouter router;
  final FallbackAppInfo fallbackAppInfo;
  final FirebaseApp firebaseApp;

  // final void Function(MetaFormService) metaFormServiceConfig;
  final String appTitle;
  final Widget Function(UserDetails profile) loggedInWidget;
  final Getter<Widget> loggedOutWidget;
  final Getter<Widget> splashPageWidget;

  AppBootstrap(
      {

      /// Plugs into [Sunny] to handle service resolution.  This allows for a
      /// swappable strategy that may use `provider` or `kiwi`.  By default,
      /// this uses the Provider package
      required this.serviceResolver,

      /// A list of main tabs.
      required this.tabs,

      /// Set this if you want to check if this version of the app is too old
      this.versionCheckSettings,

      // /// API key to be used for making calls to google
      // @required this.googleApiKey,
      //
      // /// API key to be used for making calls to google
      // @required this.googleSignInClientId,

      /// Whether the app should attempt to change brightness
      required this.isChangeBrightness,

      /// The initialized firebase app
      required this.firebaseApp,

      /// Executes code at the end of stage1, before anything Flutter related
      /// loads.  This gives a chance to do any pre-work before Widgets start flying.
      ///
      /// If the return value is a [SingleChildWidget] ([Provider]) or variant, it will
      /// be added to the top-level multi-provider
      required this.completeStage1,

      /// App identifier.  Used to prefix preferences, among other things
      required this.identifier,

      /// Builds the innerMost widget, wrapping the PlatformApp
      required this.buildPlatformApp,

      /// The fluro router to use for path resolution
      required this.router,

      /// Sends a push notification token to your server
      required this.sendDevicePushToken,

      /// Executes once within the PlatformApp component
      required this.initWithinApp,

      /// Executes once within the PlatformApp component
      required this.initWithinNavigator,

      /// Extract push message body payload
      required this.pushMessageBodyConverter,

      /// Executes when an unauthenticated exception occurs, allows you
      /// to swap for a login screen
      required this.onUnauthenticatedError,

      /// The widget to display when logged in
      required this.loggedInWidget,

      /// The widget to display when logged out
      required this.loggedOutWidget,

      /// The widget to display when logged out
      required this.splashPageWidget,

      /// The title of the app
      required this.appTitle,

      /// The appInfo to use when none could be loaded
      required this.fallbackAppInfo,

      /// The base url for this API
      required this.baseApiUrl,

      /// Runs code needed to initialize routes
      required this.initRoutes,

      /// Executes code to initialize analytics
      required this.initAnalytics,

      /// The analytics provider to use
      required this.analyticsProviders,

      /// Builds the outer widget
      required this.buildOuterContext,

      /// Builds the app themes, based on brightness
      required this.themesBuilder,

      /// Login providers to use
      required this.loginProviders,

      /// A stream of errors that bubbled up to the top level
      required this.appErrors,

      /// Loads the user profile from the server
      required this.loadUserProfile,

      /// Completes account creation
      required this.accountCompleter,

      /// Executed when the screen changes sizes
      required this.screenInit,

      /// Runs right before the user logs out.  Useful for doing any cleanup
      /// before cleaning up state
      required this.onPreLogout});

  Future run() async {
    AppBootstrap._current = this;
    if (!kIsWeb) {
      FirebaseCrashlytics.instance
          .setCrashlyticsCollectionEnabled(!kIsWeb && !kDebugMode);
    }
    // Pass all uncaught errors from the framework to Crashlytics.
    if (!kIsWeb) {
      FlutterError.onError = FirebaseCrashlytics.instance.recordFlutterError;
    }

    runZonedGuarded(() async {
      await Firebase.initializeApp();
      final appInfo = await onAppInit(Level.INFO, fallbackAppInfo);

      /// Set the resolver
      sunny.core.resolver = serviceResolver;
      _mainLog.info("---------------------------------------- ");
      _mainLog.info("---------------------------------------- ");
      _mainLog.info("");
      _mainLog.info("${appInfo.appName}");
      _mainLog.info("Package: ${appInfo.packageName}");
      _mainLog.info("Version: ${appInfo.appVersion}");
      if (appInfo.buildNumber.isNotNullOrBlank) {
        _mainLog.info("Package: ${appInfo.buildNumber}");
      }
      _mainLog.info("");
      _mainLog.info("---------------------------------------- ");
      _mainLog.info("---------------------------------------- ");
      var initUser = await fb.FirebaseAuth.instance
          .authStateChanges()
          .first
          .timeout(1.seconds, onTimeout: () {
        _mainLog
            .info("Timed out after 1 second waiting for initial firebase auth");
        return null;
      });

      _mainLog.info("");
      _mainLog.info("Starting user: ${initUser?.displayName ?? 'None'}");

      /// Initializes the sunny routing, a combination of Router 2.0 and Fluro
      final routes = await initRoutes();
      SunnyRouting.router = router;

      _mainLog.info("Routes initialized:");
      for (var o in routes) {
        _mainLog.info(" - $o");
      }

      /// This takes a few seconds, but we do it way up here so we can sit on
      /// the landing screen longer.
      UserDetails? initProfile;
      if (initUser != null) {
        FirebaseApiAuth.user = initUser;
        try {
          //AccountsApi(apiClient).profile();
          initProfile = await loadUserProfile(initUser);
        } on ApiResponseException catch (e) {
          if (e.statusCode == kPreconditionRequired ||
              e.statusCode == kUnauthorized ||
              e.statusCode == kForbidden) {
            /// Leave this alone.  It just means that we need to
            /// finish account creation
            _mainLog.warning(
                "Found a firebase token, but couldn't authenticate against the server with it");
            await fb.FirebaseAuth.instance.signOut();
            // Even though we appear to be authenticated, let's clear the state
            initUser = null;
            FirebaseApiAuth.user = null;
          } else {
            rethrow;
          }
        }
      }

      _mainLog.info("Fetching device info...");
      DeviceInfo _deviceInfo;
      try {
        _deviceInfo = await infoX.loadDeviceInfo().timeout(1.second);
      } catch (e) {
        _mainLog.info("Problem getting device info: $e");
        _deviceInfo = DeviceInfo(
          geo: null,
          isSimulator: false,
          softwareVersion: "unknown",
          locale: "en-US",
          language: "en",
          deviceType: "unknown",
          deviceModel: "unknown",
          deviceBrand: "unknown",
          ipAddress: "unknown",
          deviceId: uuid(),
          software: "unknown",
        );
      }
      _mainLog.info("Fetched device: ${_deviceInfo.deviceType}");

      final userPrefs = UserPreferencesService(identifier);
      try {
        await analytics.initialize(
          _deviceInfo,
          appInfo,
          userPrefs,
          provide: analyticsProviders,
        );
        await initAnalytics(analytics);
        _mainLog.info("Initialized Firebase Analytics");
      } catch (e) {
        _mainLog.warning("Unable to initialize Firebase analytics: $e");
      }

      final authState = AuthState(
        loadUserProfile,
        initUser,
        initProfile,
        providers: loginProviders,
        signupCompleter: accountCompleter,
        appErrors: appErrors,
      );

      if (onPreLogout != null) {
        authState.onPreLogout(onPreLogout!);
      }

      this.lifecycleInit = LifecycleInit(
        _deviceInfo,
        authState,
        analytics,
        Firebase.app(),
        router,
        appInfo,
        userPrefs,
        null,
      );

      var appContextEvents = AppContextEvents();

      /// Register app events stream
      serviceResolver.registerBuilder(null, (context) => appContextEvents);
      if (initWithinNavigator != null) {
        appContextEvents.doWithAppContext(initWithinNavigator!);
      }
      final res = await completeStage1?.call(lifecycleInit);

      runApp(
        MultiProvider(
          key: Key("mainTop"),
          providers: [
            Provider.value(value: this, updateShouldNotify: never),
            Provider.value(value: _deviceInfo, updateShouldNotify: never),
            Provider.value(value: appInfo, updateShouldNotify: never),
            Provider.value(value: userPrefs, updateShouldNotify: never),
            Provider.value(value: lifecycleInit, updateShouldNotify: never),
            // Provider.value(value: help, updateShouldNotify: never),
            Provider.value(value: router, updateShouldNotify: never),
            Provider.value(
                value: lifecycleInit!.authState, updateShouldNotify: never),
            Provider<IAuthState>.value(
                value: lifecycleInit!.authState, updateShouldNotify: never),
            if (res is SingleChildWidget) res,
            FutureProvider<FirebaseApp>(
              initialData: Firebase.app(),
              create: (_) async {
                return await Firebase.initializeApp();
              },
              catchError: (Object err, stack) {
                print(err);
                print(stack);

                throw err;
              },
              lazy: false,
            ),
          ],
          child: buildOuterContext(
            AppWidgets(
              appInfo: appInfo,
              isChangeBrightness: isChangeBrightness,
            ),
          ),
        ),
      );
    }, (obj, stackTrace) {
      appErrorStream.add(ErrorStack(obj, stackTrace));
      _mainLog.severe("Error in app: $obj", obj, stackTrace);

      analytics.encounteredError(obj, stackTrace, "App catchall");
    });
  }
}

bool notEquals<T>(T a, T b) => a != b;

bool whenNotNull<T>(T a, T b) => a == null && b != null;

bool never<T>(T a, T b) => false;

class AppWidgets extends StatefulWidget {
  final AppInfo appInfo;
  final bool isChangeBrightness;

  const AppWidgets({
    Key? key,
    required this.appInfo,
    this.isChangeBrightness = true,
  }) : super(key: key);

  @override
  _AppWidgetsState createState() => _AppWidgetsState();
}

typedef Future<T?> DeepLinkHandler<T>(BuildContext context);

class _AppWidgetsState extends BaseState<AppWidgets>
    with WidgetsBindingObserver {
  LayoutInfo? _layoutInfo;
  PlatformStyle? _targetPlatform;
  late Size _screenSize;
  DeviceInfo? _deviceInfo;
  Brightness? _brightness;
  AccountCompleter? signupCompleter;

  DateTime? _lastVersionNag;
  bool _forceVersion = false;

  @override
  void didChangePlatformBrightness() {
    final brightness = WidgetsBinding.instance!.window.platformBrightness;
    if (brightness != _brightness) {
      _mainLog.info("Brightness changed: $_brightness");
      setState(() {});
    }
  }

  @override
  void didChangeMetrics() {
    final size = WidgetsBinding.instance!.window.physicalSize /
        WidgetsBinding.instance!.window.devicePixelRatio;

    if (size != _layoutInfo!.screenSize) {
      setState(() {
        _screenSize = size;
        _resync();
      });
    }
  }

  void _resync() {
    final screenType = getDeviceType(_screenSize);
    final _target = _calculatePlatformStyle(screenType);
    _layoutInfo = LayoutInfo.ofSize(
        _screenSize, screenType, _target.first, _target.second);

    _mainLog.info("Resynced platform info: ");
    _mainLog.info("  - style: ${_layoutInfo!.platformStyle}");
    _mainLog.info("  - target: ${_layoutInfo!.targetPlatform}");
    _mainLog.info("  - size: ${_layoutInfo!.screenSize}");
    _mainLog.info("  - screenType: ${_layoutInfo!.screenType}");
  }

  @override
  void initState() {
    super.initState();

    _deviceInfo = context.get();
    assert(_deviceInfo != null);
    WidgetsBinding.instance!.addObserver(this);
    _screenSize = WidgetsBinding.instance!.window.physicalSize /
        WidgetsBinding.instance!.window.devicePixelRatio;
    _brightness = WidgetsBinding.instance!.window.platformBrightness;
    _resync();
    _checkAppVersion();
    _registerNavContext();
  }

  Tuple<PlatformStyle, TargetPlatform> _calculatePlatformStyle(
      DeviceScreenType type) {
    final platform = defaultTargetPlatform;
    if (kIsWeb) {
      /// In the case of web, we still want our app to respond accordingly
      final deviceBrand = _deviceInfo?.deviceBrand?.toLowerCase();
      if (deviceBrand == "apple") {
        switch (type) {
          // ignore: deprecated_member_use
          case DeviceScreenType.Desktop:
          case DeviceScreenType.desktop:
            return PlatformStyle.Material.to(TargetPlatform.macOS);
          default:
            return PlatformStyle.Cupertino.to(TargetPlatform.iOS);
        }
      } else {
        return PlatformStyle.Material.to(TargetPlatform.macOS);
      }
    }
    switch (platform) {
      case TargetPlatform.android:
        return PlatformStyle.Material.to(platform);
      case TargetPlatform.fuchsia:
        return PlatformStyle.Material.to(platform);
      case TargetPlatform.iOS:
        return PlatformStyle.Cupertino.to(platform);
      case TargetPlatform.linux:
        return PlatformStyle.Material.to(platform);
      case TargetPlatform.macOS:
        return PlatformStyle.Cupertino.to(platform);
      case TargetPlatform.windows:
        return PlatformStyle.Material.to(platform);
    }
  }

  void dispose() {
    WidgetsBinding.instance!.removeObserver(this);
    super.dispose();
  }

  Future _checkAppVersion() async {
    var current = AppBootstrap.current();
    final versionChecking = current.versionCheckSettings;
    final appInfo = current.lifecycleInit!.appInfo;
    if (versionChecking != null) {
      final versionInfo = await versionChecking.getMinAppVersion();
      try {
        var check = checkAppVersion(
          appInfo: appInfo,
          versionInfo: versionInfo,
        );
        _mainLog.info("Result of version check: $check");
        switch (check.result) {
          case VersionCheckResult.force:
            setState(() {
              _forceVersion = true;
            });
            return;
          case VersionCheckResult.warn:
            if (_lastVersionNag == null || _lastVersionNag!.daysAgo > 5) {
              _lastVersionNag = DateTime.now();
              _doVersionNag(appInfo);
            }
            return;
          default:
            break;
        }
      } catch (e) {
        _mainLog.warning("Unable to check version info $e", e);
      }
    }
  }

  void _doVersionNag(AppInfo appInfo) async {
    appContextEvents.doWithAppContext((context) async {
      return await showPlatformDialog<String>(
        context: context,
        barrierDismissible: false,
        builder: (BuildContext context) {
          final title = "New Update Available";
          final message =
              "There is a newer version of ${appInfo.appName} available.  We recommend that you update to the latest version.";
          final btnLabel = "Update Now";
          final btnLabelCancel = "Update Later";

          return PlatformAlertDialog(
            title: Text(title),
            content: Text(message),
            actions: [
              PlatformDialogAction(
                child: Text(btnLabel),
                onPressed: () => goToAppStore(context, appInfo),
              ),
              PlatformDialogAction(
                child: Text(btnLabelCancel),
                onPressed: () => Navigator.pop(context),
              ),
            ],
          );
        },
      );
    });
  }

  /// Whenever the app changes state, we should re-check the version status
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if (state == AppLifecycleState.resumed) {
      _checkAppVersion();
    }
  }

  @override
  Widget build(BuildContext context) {
    var app = AppBootstrap.current();
    app.screenInit.call(context, _layoutInfo!);
    final themes = context.app().themesBuilder(
        widget.isChangeBrightness ? _brightness : Brightness.light);

    if (_forceVersion == true) {
      return app.versionCheckSettings!
          .buildForcedUpgradeWidget(context, app.lifecycleInit!.appInfo);
    } else {
      return app.serviceResolver.register(
        context,
        [
          Inst.constant(themes.cardTheme, skipIfRegistered: true),
          Inst.instance(_layoutInfo),
          Inst.instance(
              widget.isChangeBrightness ? _brightness : Brightness.light),
          Inst.instance(themes.cupertinoTheme, skipIfRegistered: true),
          Inst.instance(
            themes,
            shouldNotify: (dynamic a, dynamic b) =>
                a?.brightness != b?.brightness,
          ),
        ],
        key: Key("top-providers"),
        child: Material(
          child: Theme(
            data: _brightness == Brightness.dark
                ? themes.darkTheme
                : themes.lightTheme,
            child: CupertinoTheme(
              data: themes.cupertinoTheme,
              child: themes.buildListTileTheme(
                child: SunnyApp(widget.appInfo, app: app),
              ),
            ),
          ),
        ),
      );
    }
  }

  /// Registers initializers that require a context that has access to
  /// the navigator
  void _registerNavContext() {
    appContextEvents.doWithAppContext((context) async {
      AppBootstrap.current().initWithinNavigator?.call(context);
    });
  }
}

// class SunnyRootPage extends StatefulWidget {
//   const SunnyRootPage({Key key}) : super(key: key);
//
//   @override
//   _SunnyRootPageState createState() => _SunnyRootPageState();
// }
//
// class _SunnyRootPageState extends StateObserver<SunnyRootPage>
//     with LoggingMixin {
//   bool get isLoggedIn => user != null;
//   dynamic error;
//   fb.User user;
//   StreamSubscription sub;
//   AppBootstrap _app;
//
//   /// If a deep link comes in and the user hasn't logged in, it gets captured here until
//   /// they do log in; then attempts to resolve
//   DeepLinkHandler _pendingRoute;
//
//   Stream<UserDetails> _stream;
//
//   AuthState _trippiAuth;
//
//   @override
//   void initState() {
//     super.initState();
//
//     _app = context.app();
//     _app.innerContext = context;
//     _trippiAuth = context.authState;
//     _stream = _trippiAuth.onUserChange.profile;
//
//     _trippiAuth.onUserChange.profile.listen((user) {
//       /// Applying to analytics
//       if (user != null) {
//         analytics.identify(userId: user.id, username: user.email);
//         analytics.addUserInfo(
//           name: user.name,
//           email: user.email,
//           phone: user.phone,
//         );
//       } else {
//         analytics.logLogout();
//       }
//     }).watch(this);
//
//     getStartingUrl().sanitized.nonNull((initUrl) {
//       _mainLog.info("Got app launch link: $initUrl.");
//       if (!kIsWeb) {
//         // not sure why, but on web, these links are followed automatically
//         final _pendingRoute = resolveDeepLinks(context, initUrl.toString());
//         if (_pendingRoute != null) {
//           _mainLog.info("   - $initUrl.  Resolved (${_pendingRoute != null})");
//           _pendingRoute(context);
//         } else {
//           _mainLog.warning("   - $initUrl. Could not be resolved");
//         }
//       }
//     });
//
//     getUriChangeStream().listen((link) async {
//       final sane = link.sanitized;
//       _mainLog.info("Got link change browsing: $link: sanitized to $sane");
//
//       if (!kIsWeb) {
//         final _pendingRoute = resolveDeepLinks(context, sane);
//         if (_pendingRoute != null) {
//           _mainLog.info("   - $sane.  Resolved (${_pendingRoute != null})");
//           _pendingRoute(context);
//         } else {
//           _mainLog.warning("   - $sane. Could not be resolved");
//         }
//       }
//     })?.watch(this);
//
//     /// This needs to happen far enough down the tree where there's a navigator.
//     _trippiAuth.unauthenticatedErrors.listen((_) {
//       _app.onUnauthenticatedError?.call(context);
//     }).watch(this);
//
//     _app.initWithinApp?.call(context);
//   }
//
//   @override
//   void dispose() {
//     super.dispose();
//   }
//
//   @override
//   Widget build(BuildContext context) {
//     return StreamBuilder<UserDetails>(
//       key: Key("userChanges"),
//       stream: _stream,
//       initialData: _trippiAuth.currentUser,
//       builder: (context, snap) =>
//           snap.render(context, allowNull: true, successFn: (profile) {
//         _mainLog.info("Building with ${profile?.id ?? 'no user'}");
//         return AnimatedSwitcher(
//           key: Key("loginSwitcher"),
//           child: profile == null
//               ? _app.loggedOutWidget()
//               : _app.loggedInWidget(profile),
//           duration: 400.ms,
//           switchInCurve: Curves.easeIn,
//           switchOutCurve: Curves.easeIn,
//         );
//       }, loadingFn: () {
//         return Container(color: Colors.transparent);
//       }),
//     );
//   }
// }

DeepLinkHandler? resolveDeepLinks(BuildContext context, uri) {
  final router = SunnyRouting.router;
  final match = router.matchRoute("$uri");
  if (match == null) {
    return null;
  } else {
    return (context) async {
      final result = await match.route.go(context, args: match.parameters);

      /// Inspect result?
      return result;
    };
  }
}

Locale get enUs {
  return Locale.fromSubtags(languageCode: 'en', countryCode: 'us');
}

// ignore: non_constant_identifier_names
void routeGo() {}
